package ApolloRescue.module.complex.search;

import ApolloRescue.ApolloConstants;
import ApolloRescue.module.complex.component.VisibilityInfoComponent;
import ApolloRescue.module.universal.ApolloWorld;
import ApolloRescue.module.universal.Util;
import ApolloRescue.module.universal.entities.BuildingModel;
import ApolloRescue.module.universal.entities.Path;
import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.info.AgentInfo;
import adf.agent.info.WorldInfo;
import adf.component.module.algorithm.PathPlanning;
import javolution.util.FastSet;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.awt.*;
import java.util.*;
import java.util.List;

public class HumanSearchStrategy extends AbstractSearchStrategy{

    protected Set<Building> buildings;
    private Building targetBuilding;
    private Map<Integer, EntityID> moveLogs;
    private int lastCycleExecute;
    private int thisCycleTryCount = 0;
    private Set<Polygon> polygonBlackList;
    private Set<Point> pointsBlackList;
    private Set<Building> discoveredBuildings;
    private boolean searchUnvisited;
    private Area destination;
    private int moveToPointCount;
    private VisibilityInfoComponent visibilityInfoComponent;
    private PathPlanning pathPlanning;

    public HumanSearchStrategy(ApolloWorld world, PathPlanning pathPlanning, WorldInfo worldInfo, AgentInfo agentInfo) {
        super(world, worldInfo, agentInfo);
        buildings = new FastSet<>();
        targetBuilding = null;
        this.pathPlanning = pathPlanning;
        moveLogs = new HashMap<>();
        polygonBlackList = new FastSet<>();
        lastCycleExecute = 0;
        discoveredBuildings = new HashSet<>();
        moveToPointCount = 0;
        pointsBlackList = new HashSet<>();
        this.visibilityInfoComponent = world.getComponent(VisibilityInfoComponent.class);
    }

    public Action searchPath() {
        if (ApolloConstants.DEBUG_SEARCH)
            world.printData("start path search...");
        if (path == null) {
            return null;
        }
        do {
            if (targetBuilding == null) {
                if (buildings.isEmpty()) {
                    if (ApolloConstants.DEBUG_SEARCH)
                        world.printData(path + " have no any buildings!! we should try another path...");
                    break;
                }

                Building nearestBuilding = null;
                int minTTA = Integer.MAX_VALUE;
                for (Building building : buildings) {
                    int tta = simpleTTA(building);
                    if (tta < minTTA) {
                        minTTA = tta;
                        nearestBuilding = building;
                    }
                }
                if (ApolloConstants.DEBUG_SEARCH)
                    world.printData(nearestBuilding + " is nearest building to me! assigned as target");
                targetBuilding = nearestBuilding;
            }
            Action action = searchBuilding(targetBuilding);
            if (action == null) {
                if (ApolloConstants.DEBUG_SEARCH)
                    world.printData(targetBuilding + " : CANCELED");
//                nearestBuilding.add(targetBuilding);
                buildings.remove(targetBuilding);
                targetBuilding = null;
            } else {
                return action;
            }/*else if (status.equals(SearchStatus.FINISHED)) {
                if (ApolloConstants.DEBUG_SEARCH)
                    world.printData(targetBuilding + " : FINISHED");
                discoveredBuildings.add(targetBuilding);
                buildings.remove(targetBuilding);
                targetBuilding = null;
            }*/

        } while (!buildings.isEmpty());

        return null;
    }


    public Action searchBuilding(Building building) {
        if (ApolloConstants.DEBUG_SEARCH) {
            world.printData("..............I'm going to search " + building);
        }
        if (world.getTime() == lastCycleExecute) {
            thisCycleTryCount++;
        } else {
            lastCycleExecute = world.getTime();
            thisCycleTryCount = 0;
        }

        if (building == null) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("building = null");
            }
            return null;
        }

        if (amIInLoop()) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("I'M in loop. search failed");
            }
            thisCycleTryCount = 0;
            return null;
        }
        BuildingModel buildingModel = world.getBuildingModel(building.getID());
        if (this.building == null || !this.building.getID().equals(building.getID())) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("target changed!");
            }
            blackList.clear();
            polygonBlackList.clear();
            pointsBlackList.clear();
            moveLogs.clear();
            destination = null;
            this.building = building;
        } else {
            moveLogs.put(world.getTime(), world.getSelfPosition().getID());
        }
        if (searchUnvisited && buildingModel.isVisited()) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("This building have been visited before. search finished.");
            }
            return null;
        } else if (world.isBuildingBurnt(building)) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("This building was burnt!");
            }
            return null;
        } else if (!searchUnvisited && visibilityInfoComponent.canISeeInside(buildingModel)) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("I can see inside of this building. search finished");
            }
            return null;
        }

        if (!buildingModel.isVisitable()) {//todo is visitable may not work properly... double check and test it.....
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData(buildingModel + " is not visitable");
            }
            return null;
        }

        if (destination != null) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("i am going to reach " + destination + ". this destination allocated before");
            }
            moveToPointCount = 0;

            List<EntityID> result = pathPlanning.getResult(agentInfo.getPosition(), destination.getID());
            if (result == null || result.isEmpty()) {
                blackList.add(destination);
            } else {
                return new ActionMove(result);
            }
        }

        boolean checkInside = shouldCheckInside(buildingModel);

        if (checkInside) {
            if (ApolloConstants.DEBUG_SEARCH) {
                world.printData("i should check inside of this building");
            }
            if (world.getSelfPosition().equals(building)) {
                if (ApolloConstants.DEBUG_SEARCH) {
                    world.printData("I'm in target");
                }
                return null;
            } else {
                destination = building;
                moveToPointCount = 0;
                List<EntityID> result = pathPlanning.getResult(agentInfo.getPosition(), destination.getID());
                if (result == null || result.isEmpty()) {
                    if (ApolloConstants.DEBUG_SEARCH) {
                        world.printData("i Can't Move inside of this building!");
                    }
                } else {
                    return new ActionMove(result);
                }

            }
        } else {
            destination = getDestinationArea(buildingModel);
            if (destination != null) {
                moveToPointCount = 0;
                if (ApolloConstants.DEBUG_SEARCH) {
                    world.printData("I guess i can see inside of this target out of it! destination area: " + destination);
                }
                List<EntityID> result = pathPlanning.getResult(agentInfo.getPosition(), destination.getID());
                if (result != null && !result.isEmpty()) {
                    return new ActionMove(result);
                }
                if (ApolloConstants.DEBUG_SEARCH) {
                    world.printData("failed to move in this destination. add it into blacklist!");
                }
                blackList.add(destination);
                destination = null;
                Action action = searchBuilding(building);
                if (action!=null){
                    return action;
                }
            } else {
                Pair<EntityID, Point> destinationPoint = getDestinationPoint(buildingModel);
                if (moveToPointCount >= 1) {
                    if (ApolloConstants.DEBUG_SEARCH) {
                        world.printData("Move To a point in " + buildingModel + " visitable part failed!");
                    }
                    List<EntityID> result = pathPlanning.getResult(agentInfo.getPosition(), building.getID());
                    if (result != null && !result.isEmpty()) {
                        return new ActionMove(result);
                    }

                } else {
                    if (destinationPoint.first() != null && destinationPoint.second() != null) {
                        if (ApolloConstants.DEBUG_SEARCH)
                            world.printData("Move to point...............");
                        destination = null;
                        if (world.getSelfPosition().getID().equals(destinationPoint.first())) {
                            moveToPointCount++;
                        }
                        List<EntityID> result = new ArrayList<>();
                        result.add(agentInfo.getPosition());
//                        if (result != null && !result.isEmpty()) {
                        return new ActionMove(result, destinationPoint.second().x, destinationPoint.second().y);
//                        }
//                        agent.moveToPoint(destinationPoint.first(), destinationPoint.second().x, destinationPoint.second().y);
                    } else {
                        if (ApolloConstants.DEBUG_SEARCH) {
                            world.printData("no point found to move on it. so move into building....");
                        }
                        List<EntityID> result = pathPlanning.getResult(agentInfo.getPosition(), building.getID());
                        if (result != null && !result.isEmpty()) {
                            return new ActionMove(result);
                        }
//                        agent.move(building, ApolloConstants.IN_TARGET, false);
                        if (ApolloConstants.DEBUG_SEARCH) {
                            world.printData("i can't move to this building! so set it not visitable!");
                        }
                        buildingModel.setVisitable(false);
                    }
                }
            }
        }

        destination = null;
        if (ApolloConstants.DEBUG_SEARCH) {
            world.printData("buildingSearch failed.");
        }

        return null;
    }

    private Area getDestinationArea(BuildingModel buildingModel) {
        Map<EntityID, List<Polygon>> centerVisitRoadShapes = buildingModel.getCenterVisitRoadShapes();
        Set<Area> targets = new FastSet<>();
        //first try to choose roads which visible part shape of this building contains center of it
        for (EntityID roadID : centerVisitRoadShapes.keySet()) {
            Area target = world.getEntity(roadID, Area.class);
            Pair<Integer, Integer> location = worldInfo.getLocation(target);
            Point point = new Point(location.first(), location.second());
            for (Polygon polygon : centerVisitRoadShapes.get(roadID)) {
                if (polygon.contains(point)) {
                    targets.add(target);
                }
            }
        }
//        targets.removeAll(blackList);
        if (targets.isEmpty()) {
            return null;
        }
        Area closestTarget = null;
        int minDist = Integer.MAX_VALUE;
        Point2D selfLocation = Util.getPoint(worldInfo.getLocation(agentInfo.getID()));
        Point2D areaLocation;
        for (Area area : targets) {
//            if (!blackList.contains(area)) {
            areaLocation = Util.getPoint(worldInfo.getLocation(area));
            int distance = Util.distance(selfLocation, areaLocation);
            if (distance < minDist) {
                minDist = distance;
                closestTarget = area;
            }
//            }
        }
        return closestTarget;
    }

    private Pair<EntityID, Point> getDestinationPoint(BuildingModel buildingModel) {
        EntityID destination;
        Point destinationPoint;
        Map<EntityID, List<Polygon>> centerVisitRoadShapes = buildingModel.getCenterVisitRoadShapes();
//        boolean firstTimeInLoop = true;
        Set<EntityID> roadSet = centerVisitRoadShapes.keySet();
        for (EntityID roadID : roadSet) {
            Area target = world.getEntity(roadID, Area.class);
//            if (!blackList.contains(target)) {
            if (buildingModel.getCenterVisitRoadPoints().containsKey(roadID)) {
                for (Point point : buildingModel.getCenterVisitRoadPoints().get(roadID)) {
                    if (pointsBlackList.contains(point)) {
                        continue;
                    }
                    destination = target.getID();
                    destinationPoint = new Point((int) point.getX(), (int) point.getY());

                    return new Pair<>(destination, destinationPoint);
                }
            }
        }
        if (ApolloConstants.DEBUG_SEARCH) {
            world.printData("no visitable point found in " + buildingModel);
        }
        return new Pair<>(null, null);
    }

    private boolean amIInLoop() {
        if (thisCycleTryCount >= 5) {
            if (ApolloConstants.DEBUG_SEARCH)
                world.printData("I have more than normal loop.");
            return true;
        }/* else if (thisCycleTryCount == 0) {
            int time = world.getTime();
            if (moveLogs.get(time) == moveLogs.get(time - 1)) {
                if (ApolloConstants.DEBUG_SEARCH)
                    world.printData("I'm involved in loop.");
                return true;
            }
        }*/

        return false;
    }

    public void setSearchingPath(Path searchingPath, boolean searchUnvisited) {
        this.searchUnvisited = searchUnvisited;
        if (this.path == null || !this.path.getMiddleArea().equals(path.getMiddleArea())) {
            discoveredBuildings.clear();
        }
        this.path = searchingPath;
        setShouldDiscoverBuildings();
    }

    public void setSearchUnvisited(boolean searchUnvisited) {
        this.searchUnvisited = searchUnvisited;
    }

    public void setSearchingBuilding(Building searchingBuilding) {
        this.building = searchingBuilding;
    }

    /**
     * this method check this conditions:
     * if building is too large or has unNormal shape so that center of it will not be seen from out of. Therefore we should go inside of it!
     * but if building is in fire we should not see inside of it
     *
     * @param building target building
     * @return return true if agent should check inside of this building
     */
    private boolean shouldCheckInside(BuildingModel building) {
        BuildingModel buildingModel = world.getBuildingModel(building.getID());
        if (buildingModel.getCenterVisitRoadShapes().isEmpty()) {
            if (building.getSelfBuilding().isFierynessDefined()) {
                int fieryness = building.getSelfBuilding().getFieryness();
                if (fieryness == 0 || fieryness == 4) {
                    return true;
                }
            } else {
                return true;
            }
        }
        return false;
    }


    private void setShouldDiscoverBuildings() {
        buildings.clear();
        if (path == null) {
            return;
        }
        for (Area area : path.getBuildings()) {
            BuildingModel buildingModel = world.getBuildingModel(area.getID());
//            if (!discoveredBuildings.contains(buildingModel.getSelfBuilding())) {
            if (!this.searchUnvisited || !buildingModel.isVisited()) {
                buildings.add(buildingModel.getSelfBuilding());
            }
//            }
        }
    }

    /**
     * calculate time to arrive target with euclidean distance
     *
     * @param target target area entity
     * @return needed time
     */
    protected int simpleTTA(StandardEntity target) {
        StandardEntity position = world.getSelfPosition();
        double distance = Util.distance(worldInfo.getLocation(position), worldInfo.getLocation(target));
        return (int) Math.ceil(distance / ApolloConstants.MEAN_VELOCITY_OF_MOVING);
    }
}
